/*
 * ss_err.c:
 * This file is automatically generated; please do not edit it.
 */
#ifdef __STDC__
#define NOARGS void
#else
#define NOARGS
#define const
#endif

static const char * const text[] = {
		"Subsystem aborted",
		"Version mismatch",
		"No current invocation",
		"No info directory",
		"Command not found",
		"Command line aborted",
		"End-of-file reached",
		"Permission denied",
		"Request table not found",
		"No info available",
		"Shell escapes are disabled",
		"Sorry, this request is not yet implemented",
    0
};

struct error_table {
    char const * const * msgs;
    long base;
    int n_msgs;
};
struct et_list {
    struct et_list *next;
    const struct error_table * table;
};
extern struct et_list *_et_list;

static const struct error_table et = { text, 748800L, 12 };

static struct et_list link = { 0, 0 };

void initialize_ss_error_table (NOARGS);

void initialize_ss_error_table (NOARGS) {
    if (!link.table) {
        link.next = _et_list;
        link.table = &et;
        _et_list = &link;
    }
}
